// Copyright 2008, Massachusetts Institute of Technology.
// The use of this code is permitted for research only. There is
// absolutely no warranty for this software.
//
// Author: John Lee (jjl@mit.edu)
//
// If you want a test x train matrix instead of a square matrix (i.e.,
// all-pairs similarities), use this. Given a set of test pyramids and
// a set of training pyramids, this will compute the similarities
// between each test pyramid and every training pyramid (recall that a
// "pyramid" here consists of N files where N is the number of
// channels). This will sum up the similarities over all channels,
// returning one single Matrix.
//

#include <stdio.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include "histograms/multi-resolution-histogram.h"
#include "pyramids/pyramid-matcher.h"
#include "kernel/matrix.h"
#include "util/bin-weight-scheme.h"

using namespace std;
using namespace libpmk;
using namespace libpmk_util;

void usage(const char* exec_name) {
  cerr << "Usage: " << exec_name << " output.kern train1.mrh train2.mrh ... test1.mrh test2.mrh ..\n";
  cerr << "<ouput.kern>:  Where to write unnormalized kernel to\n";
  cerr << "<train.mrh>: input pyramids\n";
  cerr << "<test.mrh>: input pyramids\n";
}

int main(int argc, char** argv) {
  if (argc < 4) {
    usage(argv[0]);
    exit(0);
  }

  string output_file(argv[1]);
  int num_channels = (argc - 2) / 2;

  Matrix matrix;
  for (int ii = 0; ii < num_channels; ++ii) {
    string train_file(argv[ii + 2]);
    string test_file(argv[ii + num_channels + 2]);
    cout << "Reading pyramids from " << train_file << "... " << flush;
    vector<MultiResolutionHistogram*>
      train(MultiResolutionHistogram::ReadFromFile(train_file.c_str()));
    cout << "done." << endl;

    cout << "Reading pyramids from " << test_file << "... " << flush;
    vector<MultiResolutionHistogram*>
      test(MultiResolutionHistogram::ReadFromFile(test_file.c_str()));
    cout << "done." << endl;

    matrix.Resize((int)test.size(), (int)train.size());

    for (int ii = 0; ii < (int)test.size(); ++ii) {
      for (int jj = 0; jj < (int)train.size(); ++jj) {
        matrix.at(ii, jj) +=
          PyramidMatcher::GetPyramidMatchSimilarity(*test[ii], *train[jj],
                                                    BIN_WEIGHT_GLOBAL);
      }
    }
    
    for (int ii = 0; ii < (int)train.size(); ++ii) {
      delete train[ii];
    }
    for (int ii = 0; ii < (int)test.size(); ++ii) {
      delete test[ii];
    }
  }
  
  cout << "Writing matrix to " << output_file << "... " << flush;
  matrix.WriteToFile(output_file.c_str());
  cout << "done." << endl;
  
  return 0;
}
